#=============================================================================
# Please provide your codes with comments/documentation where it is required. 
# You can use get_AffinityMatrix(), SNF(), plot_SNF() methods without further 
# modification. However if you need to edit these functions, please write a
# comment where the change is made. 
#
# You might also need to install additional packages (seaborn, sklearn,...)
# for the codes to run
#=============================================================================

import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.cluster import spectral_clustering
from scipy.stats import norm

def get_DistanceMatrix(dataset): 
    '''
    Return a distance matrix from a dataset
    
    Please provide your code below
    '''

def get_AffinityMatrix(distance_matrix, KNN = 5):
    '''
    Return an affinity matrix from a distance matrix
    '''
    K_sorted_distance = np.sort(distance_matrix, axis = 0).T[:, 1:(KNN+1)]
    K_average_distance = K_sorted_distance.mean(axis = 1)
    
    sigma = ((K_average_distance + K_average_distance.T)/2)/3 * 2 + distance_matrix/3
    densities = norm.pdf(distance_matrix, loc = 0, scale = sigma)
    AM = (densities + densities.T)/2
    return(AM)

def get_InitialTransitionMatrix(AM, K):
    '''
    Return initial transition matrix from an affinity matrix 
    
    Please provide your code below
    '''

def perform_SimilarityNetworkFusion(A, S, t):
    '''
    Perform SNF on a set of AffinityMatrices (A) and 
    a set of InitialLocalTransitionMatrices (S) for t steps
    Return the final updated list of AffinityMatrices
    
    Please provide your code below
    '''

def get_FusedAffinityMatrix(A_):
    '''
    Return a Fused Affinity Matrix from the finally updated 
    set of AffinityMatrices
    
    Please provide your code below
    '''
    
def SNF(A, K, t):
    '''Normalize the matrix'''
    for i in range(len(A)):
        A[i] = A[i] / np.sum( A[i] , axis=1, keepdims=True)
    
    '''Get set of InitialTransitionMatrices S from set of AffinityMatrices A'''
    S = []
    for AM in A:
        S.append(get_InitialTransitionMatrix(AM))
        
    '''Perform SimilarityNetworkFusion for t = 200 steps'''
    A_ = perform_SimilarityNetworkFusion(A, S, t)
    
    '''Compute the FusedAffinityMatrix'''
    FAM = get_FusedAffinityMatrix(A_)
    
    '''Normalize the matrix and maintain diagonal similarity'''
    FAM = FAM / np.sum(FAM, axis=1, keepdims=True)
    FAM = (FAM + FAM.T + np.eye(len(FAM))) / 2
    
    return(FAM)


def plot_SNF(FAM, n_clusters = 5):
    '''
    Plot fused network as heatmap
    '''
    np.fill_diagonal(FAM, np.median(FAM))
    FAM = FAM + FAM.T
    
    clusters = pd.Index(spectral_clustering(FAM, n_clusters = n_clusters, random_state = 0))
    sorted_clusters_idx = sorted(range(len(clusters)), key=lambda k: clusters[k])
    FAM = pd.DataFrame(FAM)
    FAM = FAM.reindex(sorted_clusters_idx, axis = 0)
    FAM = FAM.reindex(sorted_clusters_idx, axis = 1) 
    clusters = clusters[sorted_clusters_idx]
    row_pal = dict(zip(clusters.unique(), ["salmon","papayawhip","plum","wheat","chocolate"]))
    row_col = clusters.map(row_pal)
    sns.clustermap(FAM,
               row_cluster = None, col_cluster = None,
               row_colors = row_col, col_colors = row_col,
               linewidths=0)
    






           
